from collections.abc import Sequence
from pathlib import Path

import numpy as np
import scipy
from factrainer.base.dataset import BaseDatasetEqualityChecker

import lightgbm as lgb
from lightgbm.basic import (
    _LGBM_CategoricalFeatureConfiguration,
    _LGBM_FeatureNameConfiguration,
    _LGBM_GroupType,
    _LGBM_InitScoreType,
    _LGBM_LabelType,
    _LGBM_PositionType,
    _LGBM_WeightType,
)
from lightgbm.compat import (  # type: ignore
    dt_DataTable,
    pa_Array,
    pa_ChunkedArray,
    pa_Table,
    pd_DataFrame,
    pd_Series,
)

from .type import LgbDataType


class LgbDataEqualityChecker(BaseDatasetEqualityChecker[LgbDataType]):
    def check(self, left: LgbDataType, right: LgbDataType) -> bool:
        match left:
            case str():
                raise NotImplementedError
            case Path():
                raise NotImplementedError
            case np.ndarray():
                return (
                    np.array_equal(left, right) if type(right) is np.ndarray else False
                )
            case pd_DataFrame():
                return left.equals(right) if type(right) is pd_DataFrame else False
            case dt_DataTable():
                raise NotImplementedError
            case scipy.sparse.spmatrix():
                raise NotImplementedError
            case list():
                raise NotImplementedError
            case Sequence():
                raise NotImplementedError
            case pa_Table():
                raise NotImplementedError
            case _:
                raise TypeError("Invalid type")


class LgbLabelEqualityChecker(BaseDatasetEqualityChecker[_LGBM_LabelType | None]):
    def check(
        self, left: _LGBM_LabelType | None, right: _LGBM_LabelType | None
    ) -> bool:
        match left:
            case list():
                return left == right if type(right) is list else False
            case np.ndarray():
                return (
                    np.array_equal(left, right) if type(right) is np.ndarray else False
                )
            case pd_Series():
                return left.equals(right) if type(right) is pd_Series else False
            case pd_DataFrame():
                return left.equals(right) if type(right) is pd_DataFrame else False
            case pa_Array():
                raise NotImplementedError
            case pa_ChunkedArray():
                raise NotImplementedError
            case None:
                return right is None
            case _:
                raise TypeError("Invalid type")


class LgbWeightEqualityChecker(BaseDatasetEqualityChecker[_LGBM_WeightType | None]):
    def check(
        self, left: _LGBM_WeightType | None, right: _LGBM_WeightType | None
    ) -> bool:
        match left:
            case np.ndarray():
                return (
                    np.array_equal(left, right) if type(right) is np.ndarray else False
                )
            case pd_Series():
                return left.equals(right) if type(right) is pd_Series else False
            case None:
                return right is None
            case _:
                raise TypeError("Invalid type")


class LgbInitScoreEqualityChecker(
    BaseDatasetEqualityChecker[_LGBM_InitScoreType | None]
):
    def check(
        self, left: _LGBM_InitScoreType | None, right: _LGBM_InitScoreType | None
    ) -> bool:
        match left:
            case list():
                raise NotImplementedError
            case np.ndarray():
                return (
                    np.array_equal(left, right) if type(right) is np.ndarray else False
                )
            case pd_DataFrame():
                return left.equals(right) if type(right) is pd_DataFrame else False
            case pd_Series():
                return left.equals(right) if type(right) is pd_Series else False
            case pa_Table():
                raise NotImplementedError
            case pa_Array():
                raise NotImplementedError
            case pa_ChunkedArray():
                raise NotImplementedError
            case None:
                return right is None
            case _:
                raise TypeError("Invalid type")


class LgbGroupEqualityChecker(BaseDatasetEqualityChecker[_LGBM_GroupType | None]):
    def check(
        self, left: _LGBM_GroupType | None, right: _LGBM_GroupType | None
    ) -> bool:
        match left:
            case list():
                raise NotImplementedError
            case np.ndarray():
                return (
                    np.array_equal(left, right) if type(right) is np.ndarray else False
                )
            case pd_Series():
                return left.equals(right) if type(right) is pd_Series else False
            case pa_Array():
                raise NotImplementedError
            case pa_ChunkedArray():
                raise NotImplementedError
            case None:
                return right is None
            case _:
                raise TypeError("Invalid type")


class LgbFeatureNameEqualityChecker(
    BaseDatasetEqualityChecker[_LGBM_FeatureNameConfiguration]
):
    def check(
        self,
        left: _LGBM_FeatureNameConfiguration,
        right: _LGBM_FeatureNameConfiguration,
    ) -> bool:
        match left:
            case list() | "auto":
                return left == right
            case _:
                raise TypeError("Invalid type")


class LgbCategoricalFeatureEqualityChecker(
    BaseDatasetEqualityChecker[_LGBM_CategoricalFeatureConfiguration]
):
    def check(
        self,
        left: _LGBM_CategoricalFeatureConfiguration,
        right: _LGBM_CategoricalFeatureConfiguration,
    ) -> bool:
        return left == right


class LgbPositionEqualityChecker(BaseDatasetEqualityChecker[_LGBM_PositionType | None]):
    def check(
        self, left: _LGBM_PositionType | None, right: _LGBM_PositionType | None
    ) -> bool:
        return left == right


class LgbDatasetEqualityChecker(BaseDatasetEqualityChecker[lgb.Dataset | None]):
    def check(self, left: lgb.Dataset | None, right: lgb.Dataset | None) -> bool:
        if left is None:
            return right is None
        if type(right) is not lgb.Dataset:
            return False
        return all(
            [
                LgbDataEqualityChecker().check(left.data, right.data),
                LgbLabelEqualityChecker().check(left.label, right.label),
                LgbDatasetEqualityChecker().check(left.reference, right.reference),
                LgbWeightEqualityChecker().check(left.weight, right.weight),
                LgbGroupEqualityChecker().check(left.group, right.group),
                LgbInitScoreEqualityChecker().check(left.init_score, right.init_score),
                LgbFeatureNameEqualityChecker().check(
                    left.feature_name, right.feature_name
                ),
                LgbCategoricalFeatureEqualityChecker().check(
                    left.categorical_feature, right.categorical_feature
                ),
                left.params == right.params,
                left.free_raw_data == right.free_raw_data,
                LgbPositionEqualityChecker().check(left.position, right.position),
            ]
        )



from abc import ABC, abstractmethod
from typing import Self

from pydantic import BaseModel, ConfigDict

from .dataset import BaseDataset, Prediction
from .raw_model import RawModel


class BaseTrainConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BasePredictConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseLearner[T: BaseDataset, U: RawModel, V: BaseTrainConfig](ABC):
    @abstractmethod
    def train(self, train_dataset: T, val_dataset: T | None, config: V) -> U:
        raise NotImplementedError


class BasePredictor[T: BaseDataset, U: RawModel, W: BasePredictConfig](ABC):
    @abstractmethod
    def predict(self, dataset: T, raw_model: U, config: W | None) -> Prediction:
        raise NotImplementedError


class BaseMlModelConfig[
    T: BaseDataset,
    U: RawModel,
    V: BaseTrainConfig,
    W: BasePredictConfig,
](BaseModel):
    learner: BaseLearner[T, U, V]
    predictor: BasePredictor[T, U, W]
    train_config: V
    pred_config: W | None = None
    model_config = ConfigDict(arbitrary_types_allowed=True)


class MlModelConfig[
    T: BaseDataset,
    U: RawModel,
    V: BaseTrainConfig,
    W: BasePredictConfig,
](BaseMlModelConfig[T, U, V, W], ABC):
    @classmethod
    @abstractmethod
    def create(cls, train_config: V, pred_config: W | None = None) -> Self:
        raise NotImplementedError



from collections.abc import Sequence
from pathlib import Path

import pytest

THIS_DIR = Path(__file__).parent


def pytest_collection_modifyitems(
    session: pytest.Session, config: pytest.Config, items: Sequence[pytest.Item]
) -> None:
    for item in items:
        if THIS_DIR in Path(item.fspath).parents:
            item.add_marker(pytest.mark.xgboost)



from inspect import Parameter, Signature, signature
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import lightgbm as lgb
import pytest


def test_train_signature() -> None:
    expected = Signature(
        [
            Parameter(
                "params", Parameter.POSITIONAL_OR_KEYWORD, annotation=Dict[str, Any]
            ),
            Parameter(
                "train_set", Parameter.POSITIONAL_OR_KEYWORD, annotation=lgb.Dataset
            ),
            Parameter(
                "num_boost_round",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=int,
                default=100,
            ),
            Parameter(
                "valid_sets",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[List[lgb.Dataset]],
                default=None,
            ),
            Parameter(
                "valid_names",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[List[str]],
                default=None,
            ),
            Parameter(
                "feval",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[
                    Union[
                        lgb.engine._LGBM_CustomMetricFunction,
                        List[lgb.engine._LGBM_CustomMetricFunction],
                    ]
                ],
                default=None,
            ),
            Parameter(
                "init_model",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[Union[str, Path, lgb.Booster]],
                default=None,
            ),
            Parameter(
                "keep_training_booster",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=bool,
                default=False,
            ),
            Parameter(
                "callbacks",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[List[Callable]],  # type: ignore
                default=None,
            ),
        ],
        return_annotation=lgb.Booster,
    )
    actual = signature(lgb.train)

    assert actual == expected


@pytest.mark.skip
def test_lgbm_train_data_type() -> None:
    # actual = _LGBM_TrainDataType.__args__
    raise NotImplementedError



import pytest
from sklearn.datasets import fetch_california_housing
from sklearn.utils._bunch import Bunch


@pytest.fixture(scope="package")
def _california_housing_data() -> Bunch:
    return fetch_california_housing()



import warnings

warnings.warn(
    "The 'factrainer.sklearn' module is currently a placeholder and not yet implemented. "
    "Functionality will be added in a future release.",
    UserWarning,
    stacklevel=2,
)



from collections.abc import Sequence
from pathlib import Path
from typing import Any, TypeVar

import scipy
from numpy import typing as npt

from lightgbm.compat import (  # type: ignore
    dt_DataTable,
    pa_Array,
    pa_ChunkedArray,
    pa_Table,
    pd_DataFrame,
    pd_Series,
)

type LgbParams = dict[str, Any] | None
LgbDataType = TypeVar(
    "LgbDataType",
    str,
    Path,
    npt.NDArray[Any],
    pd_DataFrame,
    dt_DataTable,
    scipy.sparse.spmatrix,
    Sequence[Any],
    list[Sequence[Any]],
    list[npt.NDArray[Any]],
    pa_Table,
)

LgbLabelType = TypeVar(
    "LgbLabelType",
    list[float],
    list[int],
    npt.NDArray[Any],
    "pd_Series[Any]",
    pd_DataFrame,
    pa_Array,
    pa_ChunkedArray,
)

LgbWeightType = TypeVar(
    "LgbWeightType",
    list[float],
    list[int],
    npt.NDArray[Any],
    "pd_Series[Any]",
    pa_Array,
    pa_ChunkedArray,
)

LgbInitScoreType = TypeVar(
    "LgbInitScoreType",
    list[float],
    list[list[float]],
    npt.NDArray[Any],
    "pd_Series[Any]",
    pd_DataFrame,
    pa_Table,
    pa_Array,
    pa_ChunkedArray,
)

LgbGroupType = TypeVar(
    "LgbGroupType",
    list[float],
    list[int],
    npt.NDArray[Any],
    "pd_Series[Any]",
    pa_Array,
    pa_ChunkedArray,
)

LgbPositionType = TypeVar(
    "LgbPositionType",
    npt.NDArray[Any],
    "pd_Series[Any]",
)



from typing import Any, Self

import numpy as np
import numpy.typing as npt
from factrainer.base.config import (
    BaseLearner,
    BaseMlModelConfig,
    BasePredictConfig,
    BasePredictor,
    BaseTrainConfig,
)
from factrainer.base.dataset import IndexableDataset, Prediction
from factrainer.base.raw_model import RawModel
from joblib import Parallel, delayed

from .dataset import IndexedDatasets
from .raw_model import CvRawModels


class CvLearner[T: IndexableDataset, U: RawModel, V: BaseTrainConfig](
    BaseLearner[IndexedDatasets[T], CvRawModels[U], V]
):
    def __init__(
        self, learner: BaseLearner[T, U, V], n_jobs: int | None = None
    ) -> None:
        self._learner = learner
        self._n_jobs = n_jobs

    def train(
        self,
        train_dataset: IndexedDatasets[T],
        val_dataset: IndexedDatasets[T] | None,
        config: V,
    ) -> CvRawModels[U]:
        if val_dataset is not None:
            models = Parallel(n_jobs=self.n_jobs)(
                delayed(self._learner.train)(train.data, val.data, config)
                for train, val in zip(train_dataset.datasets, val_dataset.datasets)
            )
        else:
            models = Parallel(n_jobs=self.n_jobs)(
                delayed(self._learner.train)(train.data, None, config)
                for train in train_dataset.datasets
            )
        return CvRawModels(models=models)

    @property
    def n_jobs(self) -> int | None:
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs: int | None) -> None:
        self._n_jobs = n_jobs


class CvPredictor[T: IndexableDataset, U: RawModel, W: BasePredictConfig](
    BasePredictor[IndexedDatasets[T], CvRawModels[U], W]
):
    def __init__(
        self, predictor: BasePredictor[T, U, W], n_jobs: int | None = None
    ) -> None:
        self._predictor = predictor
        self._n_jobs = n_jobs

    def predict(
        self, dataset: IndexedDatasets[T], model: CvRawModels[U], config: W | None
    ) -> Prediction:
        y_preds = Parallel(n_jobs=self.n_jobs)(
            delayed(self._predictor.predict)(_dataset.data, _model, config)
            for _model, _dataset in zip(model.models, dataset.datasets)
        )
        y_oof_pred = self._init_pred(len(dataset), y_preds[0])
        for y_pred, _dataset in zip(y_preds, dataset.datasets):
            y_oof_pred[_dataset.index] = y_pred
        return y_oof_pred

    def _init_pred(self, total_length: int, y_pred: npt.NDArray[Any]) -> Prediction:
        return np.empty(
            tuple([total_length] + list(y_pred.shape[1:])), dtype=y_pred.dtype
        )

    @property
    def n_jobs(self) -> int | None:
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs: int | None) -> None:
        self._n_jobs = n_jobs


class CvMlModelConfig[
    T: IndexableDataset,
    U: RawModel,
    V: BaseTrainConfig,
    W: BasePredictConfig,
](BaseMlModelConfig[IndexedDatasets[T], CvRawModels[U], V, W]):
    learner: CvLearner[T, U, V]
    predictor: CvPredictor[T, U, W]

    @classmethod
    def from_config(
        cls,
        config: BaseMlModelConfig[T, U, V, W],
        n_jobs_train: int | None = None,
        n_jobs_predict: int | None = None,
    ) -> Self:
        return cls(
            learner=CvLearner(config.learner, n_jobs_train),
            predictor=CvPredictor(config.predictor, n_jobs_predict),
            train_config=config.train_config,
            pred_config=config.pred_config,
        )

    @property
    def n_jobs_train(self) -> int | None:
        return self.learner.n_jobs

    @n_jobs_train.setter
    def n_jobs_train(self, n_jobs: int | None) -> None:
        self.learner.n_jobs = n_jobs

    @property
    def n_jobs_predict(self) -> int | None:
        return self.predictor.n_jobs

    @n_jobs_predict.setter
    def n_jobs_predict(self, n_jobs: int | None) -> None:
        self.predictor.n_jobs = n_jobs



from typing import Any

import numpy as np
import numpy.typing as npt

from ._split import _BaseKFold

class _UnsupportedGroupCVMixin: ...

class KFold(_UnsupportedGroupCVMixin, _BaseKFold):
    def __init__(
        self,
        n_splits: int = 5,
        *,
        shuffle: bool = False,
        random_state: int | None = None,
    ) -> None: ...

def train_test_split(
    X: npt.NDArray[np.number[Any]],
    y: npt.NDArray[np.number[Any]] | None = None,
    test_size: float | None = None,
    random_state: int | None = None,
) -> tuple[
    npt.NDArray[np.number[Any]],
    npt.NDArray[np.number[Any]],
    npt.NDArray[np.number[Any]],
    npt.NDArray[np.number[Any]],
]: ...



from .config import LgbModelConfig, LgbPredictConfig, LgbTrainConfig
from .dataset.dataset import LgbDataset

__all__ = [
    "LgbDataset",
    "LgbModelConfig",
    "LgbPredictConfig",
    "LgbTrainConfig",
]



from pathlib import Path
from typing import Any, Callable, Self

import scipy
from factrainer.base.config import (
    BaseLearner,
    BasePredictConfig,
    BasePredictor,
    BaseTrainConfig,
    MlModelConfig,
)
from factrainer.base.dataset import Prediction

import lightgbm as lgb
from lightgbm.engine import _LGBM_CustomMetricFunction

from .dataset.dataset import LgbDataset
from .raw_model import LgbModel


class LgbTrainConfig(BaseTrainConfig):
    params: dict[str, Any]
    num_boost_round: int = 100
    valid_names: list[str] | None = None
    feval: _LGBM_CustomMetricFunction | list[_LGBM_CustomMetricFunction] | None = None
    init_model: str | Path | lgb.Booster | None = None
    keep_training_booster: bool = False
    callbacks: list[Callable[..., Any]] | None = None


class LgbPredictConfig(BasePredictConfig): ...


class LgbLearner(BaseLearner[LgbDataset, LgbModel, LgbTrainConfig]):
    def train(
        self,
        train_dataset: LgbDataset,
        val_dataset: LgbDataset | None,
        config: LgbTrainConfig,
    ) -> LgbModel:
        return LgbModel(
            model=lgb.train(
                **dict(config),
                train_set=train_dataset.dataset,
                valid_sets=[val_dataset.dataset] if val_dataset else None,
            )
        )


class LgbPredictor(BasePredictor[LgbDataset, LgbModel, LgbPredictConfig]):
    def predict(
        self, dataset: LgbDataset, model: LgbModel, config: LgbPredictConfig | None
    ) -> Prediction:
        y_pred = model.model.predict(dataset.dataset.data)
        if isinstance(y_pred, list):
            raise NotImplementedError
        elif isinstance(y_pred, scipy.sparse.spmatrix):
            raise NotImplementedError
        return y_pred


class LgbModelConfig(
    MlModelConfig[LgbDataset, LgbModel, LgbTrainConfig, LgbPredictConfig],
):
    learner: LgbLearner
    predictor: LgbPredictor
    train_config: LgbTrainConfig
    pred_config: LgbPredictConfig | None = None

    @classmethod
    def create(
        cls, train_config: LgbTrainConfig, pred_config: LgbPredictConfig | None = None
    ) -> Self:
        return cls(
            learner=LgbLearner(),
            predictor=LgbPredictor(),
            train_config=train_config,
            pred_config=pred_config,
        )
